/***************************************************************************
 FILE: com_objects.c
 LVU : 1.0.0

 DESC:
 This contains the code for object creation, destruction, copying etc...

 Copyright (C) Neil Robertson 1999
 ***************************************************************************/


#include "globals.h"


/*** Create a new object. 1st argument is name of the class, 2nd is optional 
     name of the new object.  ***/
int com_new(int com, int pc)
{
CLASS_T cl;
OBJECT_T obj;
int ret,pc2;
char *name,*param;

if (pword[pc].end - pc < 1) return ERR_SYNTAX;

/* Get class name */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

/* Find the class */
FOR_ALL_CLASSES(cl)
	if (!strcmp(cl->name,rstack_ptr->value)) break;
if (cl==NULL) return ERR_UNDEFINED_CLASS;

/* Get optional "as" and "param" arguments */
if ((ret=get_as_param(pc2,pword[pc].end,&name,&param)) != OK) return ret;

/* Create the object and its system variables. $parent is left NULL for non
   duplicated objects */
if ((obj=create_object(cl,name))==NULL) return ERR_MALLOC;

/* Create the objects system vars */
if ((ret=create_object_system_variables(obj,NULL,param))!=OK) return ret;

/* Create program (class) variables */
if ((ret=create_object_variables(obj,cl))!=OK) return ret;

/* Create $objects entry */
if (create_object_array_entry(obj)!=OK) return ERR_INTERNAL;

/* Call the constructors of the class and its parents. */
if ((ret=call_constructors(cl,NULL,obj))!=OK) return ret;

/* Return object name */
return push_rstack(obj->name);
}



/*** Duplicate an object ***/
int com_dupl(int com, int pc)
{
OBJECT_T obj,new;
VAR_T v1,v2;
int pc2,ret;
char *name,*param;

if (pword[pc].end - pc < 1) return ERR_SYNTAX;

/* Find the object to duplicate */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2))!=OK) return ret;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
FOR_ALL_OBJECTS(obj) 
	if (!strcmp(rstack_ptr->value,obj->name)) break;
if (obj==NULL) return ERR_UNDEFINED_OBJECT;

/* Get optional "as" and "param" arguments */
if ((ret=get_as_param(pc2,pword[pc].end,&name,&param)) != OK) return ret;

/* Create the object */
if ((new=create_object(obj->class,name))==NULL) return ERR_MALLOC;

/* Create the objects system vars */
if ((ret=create_object_system_variables(new,obj->name,param))!=OK) return ret;

/* Create program (class) variables */
if ((ret=create_object_variables(new,new->class))!=OK) return ret;

/* Create $object entry */
if (create_object_array_entry(new)!=OK) return ERR_INTERNAL;

/* Copy values of original objects variables into those of the new one 
   except for the system variables. This is instead of calling the 
   constructors */
FOR_ALL_VARS(v1) {
	if (v1->owner==obj && v1->name[0]!='$') {
		FOR_ALL_VARS(v2) {
			if (v2->owner==new && !strcmp(v2->name,v1->name)) {
				if ((ret=copy_variable(v1,v2,v2->private))!=OK)
					return ret;
				}
			}
		}
	}

/* Call the duplicators of the class and its parents. */
if ((ret=call_duplicators(new->class,new))!=OK) return ret;

return push_rstack(new->name);
}



/*** Gets the "as" and "param" options for the new and dupl commands ***/
int get_as_param(int pc, int end, char **name, char **param)
{
OBJECT_T obj;
int ret,opt;

*name=NULL;
*param=NULL;

while (pc <= end) {
	if (!strcasecmp(pword[pc].word,"as")) opt=1;
	else
	if (!strcasecmp(pword[pc].word,"param")) opt=2;
	else return ERR_SYNTAX;

	if (++pc > end) return ERR_SYNTAX;

	/* Get data */
	if ((ret=push_rstack_result(&pc))!=OK) return ret;

	if (opt==1) {
		/* AS */
		if (*name) return ERR_SYNTAX;  /* Had sub command twice */
		*name=rstack_ptr->value;

		/* See if name is valid */
		if (isnull(*name) || number_type(*name)!=NOT_NUMBER)
			return ERR_INVALID_ARGUMENT;

		FOR_ALL_OBJECTS(obj) 
			if (!strcmp(obj->name,*name)) 
				return ERR_DUPLICATE_DECLARATION;

		if (!check_name(*name)) return ERR_INVALID_OBJNAME;
		}
	else {
		/* PARAM */
		if (*param) return ERR_SYNTAX;
		*param=rstack_ptr->value;
		}
	}
return OK;
}



/*** Delete a list of object(s) ***/
int com_delete(int com, int pc)
{
OBJECT_T obj;
PSTACK_T context;
int pc2,ret;
if (pword[pc].end - pc < 1) return ERR_SYNTAX;

pc2=pc+1;
while(NOT_PAST_END) {
	if ((ret=push_rstack_result(&pc2))!=OK) return ret;
	if (isnull(rstack_ptr->value) || 
	    number_type(rstack_ptr->value)!=NOT_NUMBER) 
		return ERR_INVALID_ARGUMENT;

	FOR_ALL_OBJECTS(obj) {
		if (!strcmp(obj->name,rstack_ptr->value)) {
			/* Can't delete an object if we're currently in its
			   context or we'd get dangling pointers in pstack and
			   object self ref. ie: "$this->" would stop working. */
			for(context=pstack_ptr;context;context=context->prev) {
				if (context->obj==obj) 
					return ERR_CANT_DEL_OBJ_IN_CONTEXT;
				}

			/* Call destructors first */
			if ((ret=call_destructors(obj->class,obj)) != OK)
				return ret;

			/* Destroy the object */
			if (destroy_object(obj) != OK) return ERR_INTERNAL;
			break;
			}
		}
	if (obj==NULL) return ERR_UNDEFINED_OBJECT;
	}
return OK;
}



/*** Promote an object to a higher child class ***/
int com_promote(int com, int pc)
{
OBJECT_T obj;
CLASS_T cl,old_cl;
VAR_T v;
int ret,pc2;
char *objname;

if (pword[pc].end - pc < 3) return ERR_SYNTAX;

/* Get object name */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (pc2 >= pword[pc].end) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
objname=rstack_ptr->value;

/* Check for "to" */
if (strcasecmp(pword[pc2].word,"to")) return ERR_SYNTAX;

/* Get class */
++pc2;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

FOR_ALL_CLASSES(cl)
	if (!strcmp(cl->name,rstack_ptr->value)) break;
if (cl==NULL) return ERR_UNDEFINED_CLASS;

/* Find object */
FOR_ALL_OBJECTS(obj) 
	if (!strcmp(objname,obj->name)) break;
if (obj==NULL) return ERR_UNDEFINED_OBJECT;

/* Check object belongs to a parent class of the class to promote to */
if (obj->class==cl || !is_derived_class(cl,obj->class)) 
	return ERR_OBJ_CLASS_NOT_PARENT;

/* Ok , everythings in order , lets promote it ... */
old_cl=obj->class;
obj->class=cl;

/* Create new class variables first. */
if ((ret=create_object_variables(obj,cl))!=OK) return ret;

/* Reset the $class & $promoted object system variables */
FOR_ALL_VARS(v) {
	if (v->owner==obj) {
		if (!strcmp(v->name,"$class")) set_string(&v->value,cl->name);  
		if (!strcmp(v->name,"$promoted")) set_string(&v->value,"1");
		}
	}

/* Reset $object entry */
if (create_object_array_entry(obj)!=OK) return ERR_INTERNAL;

/* Call all the constructors up to but not including the objects old class.
   This might not work properly with multiple inheritence if 2 branches
   share the same parent. */
if ((ret=call_constructors(cl,old_cl,obj)) != OK) return ret;

/* Return new class for want of anything else */
return push_rstack(cl->name);
}



/*** Create the $name, $class etc system variables for an object ***/
int create_object_system_variables(
	OBJECT_T obj, 
	char *dupl, 
	char *param)
{
VAR_T var;

if ((var=create_variable("$name",obj,NULL,NO,YES))==NULL) return ERR_MALLOC;
add_variable(var);
var->value=strdup(obj->name);

if ((var=create_variable("$class",obj,NULL,NO,YES))==NULL) return ERR_MALLOC;
add_variable(var);
var->value=strdup(obj->class->name);

if ((var=create_variable("$created",obj,NULL,NO,YES))==NULL) return ERR_MALLOC;
add_variable(var);
sprintf(text,"%u",time(0));
var->value=strdup(text);

if ((var=create_variable("$parent",obj,NULL,NO,YES))==NULL) return ERR_MALLOC;
add_variable(var);
set_string(&var->value,dupl);  /* use set_string cos dupl could be NULL */

if ((var=create_variable("$promoted",obj,NULL,NO,YES))==NULL) return ERR_MALLOC;
add_variable(var);
var->value=strdup("0");

if ((var=create_variable("$param",obj,NULL,NO,YES))==NULL) return ERR_MALLOC;
add_variable(var);
set_string(&var->value,param);

return OK;
}



/*** Create an entry in the $objects array ***/
int create_object_array_entry(OBJECT_T obj)
{
char varname[100];

sprintf(varname,"$objects[\"%s\"]",obj->name);
return set_var_value(varname,NULL,obj->class->name,YES);
}



/*** WITH command ***/
int com_with(int com, int pc)
{
OBJECT_T obj;
WITH_T with;
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get object name */
pc2=pc+1;

/* Get object name */
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

/* Get object pointer */
FOR_ALL_OBJECTS(obj) 
	if (!strcmp(rstack_ptr->value,obj->name)) break;
if (obj==NULL) return ERR_UNDEFINED_OBJECT;

/* Get with block */
for(with=first_with;with;with=with->next)
	if (with->start==pc) break;
if (with==NULL) return ERR_INTERNAL;

/* Push onto wstack */
return push_wstack(with,obj);
}



/*** WITHOUT command ***/
int com_without(int com, int pc)
{
if (pword[pc].end != pc) return ERR_SYNTAX;

/* Don't flag an error if wstack_ptr is null since a loop may have 
   exited the stack for us. See com_loopend(). */
if (wstack_ptr && wstack_ptr->context!=pstack_ptr)
	return ERR_UNEXPECTED_COMMAND;

pop_wstack();
return OK;
}

