/***************************************************************************
 FILE: com_processes.c
 LVU : 1.0.0

 DESC:
 This contains code to run and interface with other processes on the 
 computer plus fork and exec'ing.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"

#define FREE_EXEC_ARRAY()  for(i=0;args[i];++i) free(args[i]);  free(args)


/*** Execute a shell command and return a simple code ***/
int com_system(int com, int pc)
{
int ret,pc2;

if (pword[pc].end == pc) return ERR_SYNTAX;

/* Get command string */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;

/* Execute it */
sprintf(text,"%d",system(rstack_ptr->value));
return push_rstack(text);
}



/*** Create a pipe stream to another process ***/
int com_popen(int com, int pc)
{
STREAM_T new;
struct timeval tm;
union semun semarg;
pid_t pid;
int end,ret,pc2,type,i;
int p[2],sem,option,roc,secs,usecs;
char *execstr,**args;

if ((end=pword[pc].end) - pc < 3) return ERR_SYNTAX;

/* Get exec string */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (IS_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
execstr=rstack_ptr->value;

/* Noise word "to" */
if (strcasecmp(pword[pc2].word,"to")) return ERR_SYNTAX;

/* Open method */
if (++pc2 > end) return ERR_SYNTAX;
if (!strcasecmp(pword[pc2].word,"read")) type=STM_READ;
else
if (!strcasecmp(pword[pc2].word,"write")) type=STM_WRITE;
else
if (!strcasecmp(pword[pc2].word,"readwrite")) type=STM_READWRITE;
else
return ERR_SYNTAX;

/* Get options */
roc=NO;
option=0;
while (++pc2 <= end) {
	if (!strcasecmp(pword[pc2].word,"dowait")) {
		if (option==2) return ERR_MUTUALLY_EXCLUSIVE;

		/* Get wait time */
		if (pc2++ == end) return ERR_SYNTAX;
		if ((ret=push_rstack_result(&pc2)) != OK) return ret;	
		--pc2;
		if (number_type(rstack_ptr->value) == NOT_NUMBER)
			return ERR_INVALID_ARGUMENT;

		secs=atoi(rstack_ptr->value);
		usecs=get_usecs(rstack_ptr->value);
		
		option=1;
		}
	else
	if (!strcasecmp(pword[pc2].word,"dosig")) {
		if (option==1) return ERR_MUTUALLY_EXCLUSIVE;
		option=2;
		}
	else 
	if (!strcasecmp(pword[pc2].word,"roc")) roc=YES; /* reap on close */
	else
	return ERR_SYNTAX; 
	}

/* Get args from exec string */
if ((ret=create_exec_array(execstr,&args)) != OK) return ret;

/* Create a semaphore set so we can tell if child process worked ok */
if (option==1) {
	if ((sem=semget(IPC_PRIVATE,1,IPC_CREAT | IPC_EXCL | 0600)) < 0)
		return ERR_CANT_CREATE_SEMAPHORE;
	semarg.val=0;
	}

/* Create a pipe, fork & exec the process */
if (pipe(p) < 0) {
	FREE_EXEC_ARRAY();  
	semctl(sem,0,IPC_RMID,semarg);	
	return ERR_CANT_CREATE_PIPE;
	}

switch(pid=fork()) {
	case -1: 
	FREE_EXEC_ARRAY();  return ERR_CANT_FORK;	

	case 0:
	/* Redirect new processes stdin, out & err down the pipe */
	dup2(p[0],0);
	dup2(p[1],1);
	dup2(p[1],2);
	execvp(args[0],args);

	/* If we've failed set semaphore or send sigpipe or do nothing 
 	   depending on option */
	switch(option) {
		case 1:
		semarg.val=1;
		semctl(sem,0,SETVAL,semarg);
		exit(0);

		case 2:
		kill(getppid(),SIGPIPE); 
		}
	exit(0);
	}
FREE_EXEC_ARRAY();

/* Set this side of the pipe to non blocking. Close any side of pipe thats
   not going to be used so we can get an EOF */
if (type==STM_WRITE) close(p[0]);
else fcntl(p[0],F_SETFL,O_NDELAY);
if (type==STM_READ) close(p[1]);
else fcntl(p[1],F_SETFL,O_NDELAY);

/* Idle for the given wait time then check the semaphore */
if (option==1) {
	/* Use select to sleep since usleep() not available on all systems */
	tm.tv_sec=secs;
	tm.tv_usec=usecs;
	select(0,NULL,0,0,&tm);

	if (semctl(sem,0,GETVAL,semarg)) {
		semctl(sem,0,IPC_RMID,semarg);  return ERR_EXEC_FAILED;
		}

	/* Delete semaphore */
	semctl(sem,0,IPC_RMID,semarg);
	}

/* Free array & create pipe stream */
if ((ret=create_stream(execstr,ST_PIPE,type,0,-1,pid,&new))!=OK) return ret;
new->fd=p[0];
new->fd2=p[1];
if (roc) new->pid=pid;
	
return push_rstack(new->name);
}



/*** Wait for a child process to exit ***/
int com_wait(int com, int pc)
{
int ret,pc2,options,status,i;
pid_t pid;

/* Get optional pid and options and do a waitpid */
if (pword[pc].end > pc) {
	pc2=pc+1;
	if ((ret=push_rstack_result(&pc2)) != OK) return ret;
	if (number_type(rstack_ptr->value) != INTEGER ||
	    (pid=(pid_t)atoi(rstack_ptr->value)) < 1)
		return ERR_INVALID_ARGUMENT;

	options=0;
	if (NOT_PAST_END) {
		if (pword[pc].end - pc2 > 1) return ERR_SYNTAX;

		for(i=pc2;i <= pword[pc].end;++i) {
			if (!strcasecmp(pword[i].word,"nohang")) 
				options |= WNOHANG;
			else
			if (!strcasecmp(pword[i].word,"untraced")) 
				options |= WUNTRACED;
			else
			if (!strcasecmp(pword[i].word,"continued")) 
#ifdef SOLARIS
				options |= WCONTINUED;
#else
				return ERR_UNAVAILABLE;
#endif
			else
			if (!strcasecmp(pword[i].word,"nowait")) 
#ifdef SOLARIS
				options |= WNOWAIT;
#else
				return ERR_UNAVAILABLE;
#endif
			else
			return ERR_SYNTAX;
			}
		}

	/* Do the wait */
	if ((pid=waitpid(pid,&status,options)) == -1) return ERR_WAIT;
	}
else
if ((pid=wait(&status)) == -1) return ERR_WAIT;

/* Return the pid and the various statuses as a list */
sprintf(text,"%d %d %d %d %d %d %d",
	(int)pid,
	WIFEXITED(status),
	WEXITSTATUS(status),
	WIFSIGNALED(status),
	WTERMSIG(status),
	WIFSTOPPED(status),
	WSTOPSIG(status));

return push_rstack(text);
}



/*** Fork a copy of the interpreter ***/
int com_fork(int com, int pc)
{
pid_t pid;

if (pword[pc].end != pc) return ERR_SYNTAX;

switch(pid=fork()) {
	case -1: return ERR_CANT_FORK;

	/* Child */
	case 0: 
	/* Reset the pid system vars */
	sprintf(text,"%d",getpid());
	set_var_value("$pid",pstack_ptr,text,YES);

	sprintf(text,"%d",getppid());
	set_var_value("$ppid",pstack_ptr,text,YES);

	return push_rstack("0");
	}

/* Parent */
sprintf(text,"%d",pid);
return push_rstack(text);
}



/*** Exec a new process ***/
int com_exec(int com, int pc)
{
int ret,pc2,i;
char **args;

if (pword[pc].end - pc == pc) return ERR_SYNTAX;

/* Get system command and its arguments */
pc2=pc+1;
if ((ret=push_rstack_result(&pc2)) != OK) return ret;
if (NOT_PAST_END) return ERR_SYNTAX;
if (isnull(rstack_ptr->value)) return ERR_INVALID_ARGUMENT;
if ((ret=create_exec_array(rstack_ptr->value,&args)) != OK) return ret;

/* Do the exec */
execvp(args[0],args);

/* If we get here it failed */
FREE_EXEC_ARRAY();
return ERR_EXEC_FAILED;
}



/*** Create an array out of a string. Used by com_popen and com_exec ***/
int create_exec_array(char *str, char ***args)
{
char *s,*s2,c;
int i;

/* Count how many arguments we have then create 1st level of array */
for(s=str,i=0;*s;) {
	while(*s < 33 && *s) ++s;
	if (!*s) break;
	++i;
	while(*s > 32) ++s;
	}
if (!i) return ERR_INVALID_ARGUMENT;
if ((*args=(char **)malloc(i * sizeof(char *))) == NULL) return ERR_MALLOC;

/* Create 2nd level of array */
for(s=str,i=0;*s;) {
	while(*s < 33 && *s) ++s;
	if (!*s) break;
	s2=s;
	while(*s2 > 32) ++s2;
	c=*s2;  *s2='\0';
	(*args)[i++]=strdup(s);
	*s2=c;  s=s2;
	}
(*args)[i]=NULL;
return OK;
}
