/***************************************************************************
 FILE: globals.h
 LVU : 1.0.0

 DESC:
 This contains the global variable definitions along with numerous macros
 and enums.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#ifdef _AIX
#include <sys/select.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#ifdef LINUX
#include <sys/ioctl.h>
#include <sys/ptrace.h>
#endif
#ifdef IRIX
#include <stropts.h>
#include <limits.h>
#endif
#include <time.h>
#include <math.h>
#include <termios.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <unistd.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>

#include "version.h"
#include "build_date.h"

#define MAX_INCLUDE_NEST 10
#define PROG_WORD_ALLOC 50;
#define SCRATCH_SIZE 1000

#define NUM_COMS 216
#define NUM_ERRORS 120
#define NUM_OPERATORS 25
#define SIGQ_SIZE 20

/* 160 seems to be a high enough number for all common errors */
#define NUM_OS_ERRORS 160

/* For the maths and drawing commands */
#define DEGS_PER_RADIAN 57.29578

#define TAB 9

#define DEFAULT_FILE_PERMS 0644 /* Default new file permissions */
#define DEFAULT_DIR_PERMS 0755 /* Default new directory permissions */

#define NULL_ELEMENT "[]"
#define REGEXP_CHARS "*?[{"

#ifndef MAIN_FILE
#define EXTERN extern
#else
#define EXTERN
#endif


/* Main structs */
EXTERN
struct program {
	char *word;
	char *filename;
	int opcode;
	int com;
	int realline;
	int progline;
	int end;

	/* Equation and boolean elements */
	int processed;
	int result_set;
	int next_pc;
	double result;
	} *pword;

EXTERN int alloced_words;
EXTERN int num_words;
EXTERN int private_member;

EXTERN
struct source_data {
	FILE *fp;
	char *filename;
	int realline;
	} ifile[MAX_INCLUDE_NEST];
	

typedef 
struct classes {
	char *name;
	u_int obj_cnt;

	int var_cnt;
	struct class_vars {
		char *name;
		int private;
		} **var;

	struct procedures *constructor;
	struct procedures *destructor;
	struct procedures *duplicator;

	int inherit_cnt;
	struct classes **inherit;
	struct classes *prev,*next;
	}* CLASS_T;

EXTERN CLASS_T first_class, last_class, curr_class;

typedef
struct objects {
	char *name;
	struct classes *class;	
	struct objects *prev,*next;
	}* OBJECT_T;

EXTERN OBJECT_T first_obj, last_obj, curr_obj;


typedef
struct vars {
	char *name;
	char *value;
	int dim;
	int private;
	int read_only;
	OBJECT_T owner;
	struct vars *refvar;
	struct pstack *context;
	struct vars *array;
	struct vars *prev, *next;
	}* VAR_T;

EXTERN VAR_T first_var, last_var, curr_var;


/* Procedures */
typedef
struct procedures {
	char *name;
	int start;
	int end;
	int private;
	int vproc;
	CLASS_T class;
	struct procedures *next;
	}* PROC_T;

EXTERN PROC_T first_proc, last_proc, curr_proc;


/* Exceptions. Commands are "catch", "catchrun" and "caught" */
typedef 
struct exceptions {
	int trapped[NUM_ERRORS];
	int start;
	int end;
	PROC_T proc;
	struct exceptions *next;
	}* EXCP_T;

EXTERN EXCP_T first_excp, last_excp, curr_excp;

/* Store positions of TRY - TRIED blocks for exceptions */
typedef
struct try {
	int start;
	int end;
	struct try *next;
	}* TRY_T;

EXTERN TRY_T first_try, last_try, curr_try;


/* WITH - WITHOUT blocks */
typedef
struct with {
	PROC_T proc;
	int start;
	int end;
	struct with *prev,*next;
	}* WITH_T;

EXTERN WITH_T first_with, last_with, curr_with;


/* Loops (including switch command) data */
typedef 
struct loops {
	int type;
	int start;
	int end;
	PROC_T proc;
	struct loops *prev, *next;
	}* LOOP_T;

EXTERN LOOP_T first_loop, last_loop;

/* Labels for gotos */
typedef
struct labels {
	char *name;
	int pos;
	PROC_T proc;
	struct labels *next;
	}* LABEL_T;

EXTERN LABEL_T first_label, last_label;

/* STREAMS */
typedef 
struct streams {
	char *name;
	char *filename;  /* Actual name of file or pipe command */
	int fd;   /* I/O fd for all streams except pipes where input only */
	int fd2;  /* used for pipes only , this is the output */
	int type;
	int open_mode;
	int pos;
	int read_cnt;
	int write_cnt;
	int block;
	int stripnl;
	int telopt_cnt;
	int rows, cols;  /* For terminal */
	pid_t pid;  /* child pid if reap on close, else 0. Pipes only */
	struct sockaddr_in net_addr;
	struct streams *prev,*next;
	}* STREAM_T;

EXTERN STREAM_T first_stream, last_stream, in_stream, out_stream;
EXTERN u_int stream_cnt;

/* Change arrays in create_stream_array_entry() if these enums are changed */
enum stream_types { ST_CONSOLE, ST_FILE, ST_SOCKET, ST_LSOCKET, ST_PIPE };
enum stream_open_mode { STM_READ, STM_WRITE, STM_APPEND, STM_READWRITE };

/* STACKS */
typedef 
struct pstack {
	PROC_T proc;
	OBJECT_T obj; /* Set if we're running in an objects procedure */
	int ret_pc; /* place to jump back to in this proc after a call */
	int catchreturn_pc[2];  /* Places to jump back to in try block */
	int catchbreak_pc;  /* End of catch-caught block */
	struct pstack_pstore **pstore;
	struct pstack *prev;
	}* PSTACK_T;

EXTERN PSTACK_T pstack_start,pstack_ptr;

/* Stores maths and boolean results in case procedure is called recursively */
struct pstack_pstore {
	int processed;
	int result_set;
	int next_pc;
	double result;
	};


typedef
struct rstack {
	char *value;
	PSTACK_T context;
	struct rstack *prev;
	}* RSTACK_T;

EXTERN RSTACK_T rstack_start,rstack_ptr;

typedef
struct lstack {
	LOOP_T loop;
	int loopcom_to;
	int loopcom_curr;
	double to_val;
	double step_val;
	VAR_T arr_index;
	VAR_T arr_value;
	VAR_T forent;
	char *forelem_list;
	char *forelem_es;
	PSTACK_T context;
	struct lstack *prev;
	}* LSTACK_T;

EXTERN LSTACK_T lstack_start,lstack_ptr;

typedef
struct wstack {
	OBJECT_T obj;
	PSTACK_T context;
	WITH_T with;
	struct wstack *prev;
	}* WSTACK_T;

EXTERN WSTACK_T wstack_start,wstack_ptr;


/* USER STACKS. These use a 2 level linked list. */
typedef
struct ustack {
	char *value;  /* Value of entry or name of stack */
	struct ustack *start, *end;  /* Start and end points for 2ndary list */
	struct ustack *prev, *next;
	} *USTACK_T;

EXTERN USTACK_T ustack_start, ustack_end;


/* SIGNALS */
typedef
struct sighandler {
	int signum;
	int siglist_index;
	PROC_T handler;
	struct sighandler *prev,*next;
	}* SIGHANDLER_T;

EXTERN SIGHANDLER_T sighandler_start, sighandler_end;
EXTERN SIGHANDLER_T sigq[SIGQ_SIZE];
EXTERN int sigq_start, sigq_next;

struct siglist_struct {
	char *name;
	int signum;
	};

#define NUM_SIGNALS 14

#ifdef MAIN_FILE
struct siglist_struct siglist[NUM_SIGNALS]={
	"INT",    SIGINT,
	"QUIT",   SIGQUIT,
	"KILL",   SIGKILL,
	"TSTP",   SIGTSTP,
	"ALARM",  SIGALRM,
	"TERM",   SIGTERM,
	"STOP",   SIGSTOP,
	"USER1",  SIGUSR1,
	"USER2",  SIGUSR2,
	"TTYIN",  SIGTTIN,
	"TTYOUT", SIGTTOU,
	"CHILD",  SIGCLD,
	"WINSIZE",SIGWINCH,
	"PIPE",   SIGPIPE
	};
#else
extern struct siglist_struct siglist[];
#endif


/* Code macros */
#define FOR_ALL_CLASSES(c)      for(c=first_class;c;c=c->next)
#define FOR_ALL_OBJECTS(o)      for(o=first_obj;o;o=o->next)
#define FOR_ALL_PROCS(p)        for(p=first_proc;p;p=p->next)
#define FOR_ALL_EXCEPTIONS(e)   for(e=first_excp;e;e=e->next)
#define FOR_ALL_VARS(v)         for(v=first_var;v;v=v->next)
#define FOR_VAR_ELEMENTS(v1,v2) for(v2=v1->array;v2;v2=v2->next)
#define FOR_ALL_LOOPS(l)        for(l=first_loop;l;l=l->next)
#define FOR_ALL_LABELS(l)       for(l=first_label;l;l=l->next)
#define FOR_ALL_STREAMS(s)      for(s=first_stream;s;s=s->next)
#define FOR_LOOP_STACK(l)       for(l=lstack_ptr;l;l=l->prev)
#define FOR_USER_STACK(u)       for(u=ustack_start;u;u=u->next)
#define FOR_ALL_SIGHANDLERS(s)  for(s=sighandler_start;s;s=s->next)

#define FREE(m) if (m) free(m)
#define NOT_PAST_END pc2 <= pword[pc].end
#define IS_PAST_END  pc2 > pword[pc].end


/* Simple variables */
EXTERN int include_level, oob_top, oob_bottom, oob_left, oob_right;
EXTERN int manual_throw;
EXTERN char text[SCRATCH_SIZE];
EXTERN char *element_seps;
EXTERN char *empty_stack_marker;


#ifdef SOLARIS
/* The structure the bsd libs require. This is set up incorrectly in
   the Solaris header files for the BSD readdir func as it contains an extra
   field which it doesn't want which in turn causes incorrect data. */
typedef struct {
	ino_t d_ino;      /* "inode number" of entry */
	off_t d_off;      /* offset of disk directory entry */
	char  d_name[1];  
	} loor_dirent;
#else
typedef struct dirent loor_dirent;
#endif

#if defined SOLARIS || defined _SEM_SEMUN_UNDEFINED
/* Seems a bit daft to me having to specify this ourselves but who knows 
   what X/OPEN were thinking */
union semun {
	int val;
	struct semid_ds *buf;
	unsigned short int *array;
	struct seminfo *__buf;  /* Linux only? */
	};
#endif


/* Function defs */

/* main.c */
void init_interpreter();
void get_command_line(int , char **);
int mainloop(int, int);
int run_command(int *);
int in_try_block(int);

/* parse.c */
void load_and_first_parse();
void add_word(int ,int ,int *, int *);
void get_include_filename(FILE *,int);
void open_file(int);
void second_parse();
int get_command_number(int);
void set_alias(int);

/* procs.c */
int parse_proc_line(int);
int parse_exception_handler(int);
EXCP_T create_exception_block();
int exception_already_handled(int);
int create_try_block(int);
int create_with_block(int);
void exit_procedure();
int get_proc(char *, PROC_T *, OBJECT_T *);

/* objects.c */
int parse_class_line(int);
CLASS_T create_class();
int inherit_class(CLASS_T, CLASS_T);
int parse_class_variable_line(int, CLASS_T, int);
OBJECT_T create_object(CLASS_T, char *);
int create_object_variables(OBJECT_T obj, CLASS_T);
int destroy_object(OBJECT_T);
int is_derived_class(CLASS_T, CLASS_T);
PROC_T get_class_proc(CLASS_T, char *);
int call_constructors(CLASS_T, CLASS_T, OBJECT_T);
int call_duplicators(CLASS_T, OBJECT_T);
int call_destructors(CLASS_T, OBJECT_T);

/* vars.c */
int parse_variable_line(int, OBJECT_T, PSTACK_T);
int check_variable_name(char *, OBJECT_T, PSTACK_T);
VAR_T create_variable(char *, OBJECT_T, PSTACK_T, int, int);
int copy_variable(VAR_T, VAR_T, int);
void add_variable(VAR_T);
void destroy_variable(VAR_T);
void remove_variable(VAR_T);
int get_var_value(char *, PSTACK_T);
int get_var_main(char *, PSTACK_T, VAR_T *, char **);
int getset_var_element(VAR_T, PSTACK_T, VAR_T *, char *, char *, int, int *);
int set_var_value(char *, PSTACK_T, char *, int);
int delete_array_entry(char *, char *);

/* loops.c */
int create_label(int);
int create_loop(int);
int terminate_loop(int);
LOOP_T get_loop(int);

/* stacks.c */
int push_rstack_result(int *);
int push_rstack(char *);
void pop_rstack();
void cleardown_rstack();
int push_pstack(PROC_T, int, OBJECT_T);
void pop_pstack();
int store_in_pstack(int);
void copy_from_pstack(int);
int push_lstack(LOOP_T);
void pop_lstack();
int push_wstack(WITH_T, OBJECT_T);
void pop_wstack();
void cleardown_lw_stacks(int);

/* signals.c */
void set_sighandler();
void set_sig_default(int);
void signal_handler(int);
int process_sigqueue(int);
void push_sigqueue(SIGHANDLER_T);
void pop_sigqueue();

/* streams.c */
int create_stream(char *, int, int, mode_t, int, pid_t, STREAM_T *);
int destroy_stream(char *);
int create_stream_array_entry(STREAM_T);
STREAM_T get_stream(char *);
void write_stream(STREAM_T, char *, int);

/* bool.c */
int evaluate_complete_boolean(int, int *);
int evaluate_compound_boolean(int, int, int *);
int evaluate_single_boolean(int, int *, int *);
int compare_strings(char *, char *);

/* regexp.c */
int is_regexp(char *);
int regexp_match(char *, char *);
int regexp_charmatch(char, char, int, char *, char *);
void regexp_get_range(char *, char **, char **);
int regexp_get_count(char *);

/* misc.c */
void error_exit(int ,int ,char *);
void system_exit(int);
int check_name(char *);
int number_type(char *);
int set_string(char **, char *);
int append_string(char **, char *);
int append_char(char **, char, int);
int isnull(char *);
int nested_call_command(int);
int is_element_sep(char, char *);
int get_usecs(char *);

/* com_vars.c */
int com_var(int, int);
int com_set(int, int);
int com_aset(int, int);
int com_wevar(int, int);
int com_arrsize(int, int);
int erase_vararray(VAR_T);

/* com_objects.c */
int com_new(int, int);
int com_dupl(int, int);
int get_as_param(int, int, char **, char **);
int com_delete(int, int);
int com_promote(int, int);
int create_object_system_variables(OBJECT_T, char *, char *);
int create_object_array_entry(OBJECT_T);
int com_with(int, int);
int com_without(int, int);

/* com_procs.c */
int com_call(int, int);
int com_return(int, int);

/* com_if.c */
int com_if(int, int *);
int com_else(int, int *);
int com_bool(int, int);

/* com_io.c */
int com_select(int, int);
int com_print(int, int);
int com_input(int, int);
int com_in_out(int, int);
int com_open(int, int);
int com_close(int, int);
int com_wind(int, int);
int com_spos(int, int);
int com_stat(int, int);
int com_ioctl(int, int);
int com_dir(int, int);
int com_chmkdf(int, int);
int com_lcm(int, int);
int copy_file(char *, char *, mode_t);
int com_rm_attrib(int, int);
int com_umask(int, int);
int get_perm_value(mode_t *);
void get_sw_pos(char *, char **, char **);
void strip_escapes(char *);

/* com_tcp.c */
int com_connect(int, int);
int com_listen(int, int);
int com_acc_gethost(int, int);
int com_getip(int, int);

/* com_loops.c */
int com_goto(int, int *);
int com_loopcom(int, int *);
int com_for(int, int *);
int com_forent(int, int *);
int com_forelem(int, int *);
int com_while(int, int *);
int com_do(int, int);
int com_whilst(int, int *);
int com_switch(int, int *);
int com_case_def(int, int);
int com_break_cont(int, int *);
int com_loopend(int, int *);
int in_loop(LOOP_T);

/* com_strings.c */
int com_addstr(int, int);
int com_strfuncs1(int, int);
int com_strfuncs2(int, int);
int com_strfuncs3(int, int);
int com_ulstr(int, int);
int com_midstr(int, int);
int com_format(int, int);
int com_oistr(int, int);
int com_replstr(int, int);

/* com_lists.c */
int com_head_tail(int, int);
int com_element(int, int);
int com_midlist(int, int);
int com_member(int, int);
int com_tidy(int, int);
int com_oielem(int,int);
int com_sizeuniq(int, int);
int com_sort(int, int);
int com_sdlist(int, int);
int com_replace(int, int);
int com_join_keys(int, int);
int com_split(int, int);

/* com_maths.c */
int com_eqn(int, int);
int evaluate_equation(int, int, int);
int com_not(int, int);
int com_incdec(int, int);
int com_random(int, int);
int com_minmax(int, int);
int com_trig(int, int);
int com_maths1(int, int);
int com_maths2(int, int);

/* com_time.c */
int com_sleep(int, int);
int com_date(int, int);

/* com_excp.t */
int com_trap(int, int);
int com_throw(int, int *);
int com_catchreturn(int, int *);
int com_catchbreak(int, int *);

/* com_terminal.c */
int com_cls(int, int);
int com_locate(int, int);
int com_oob(int, int);
int com_line(int, int);
int com_circle(int, int);
int location_visible(STREAM_T, int, int);
int com_termsize(int, int);

/* com_signals.c */
int com_signal(int, int);
int com_alarm(int, int);
int com_sendsig(int, int);

/* com_processes.c */
int com_system(int, int);
int com_popen(int, int);
int com_wait(int, int);
int com_fork(int, int);
int com_exec(int, int);
int create_exec_array(char *, char ***);

/* com_stacks.c */
int com_push(int, int);
int com_pop(int, int);
int com_wipestack(int, int);

/* com_misc.c */
int com_asc(int, int);
int com_exit(int, int);
int com_getenv(int, int);
int com_setenv(int, int);
int com_crypt(int, int);
char *loor_crypt(char *, char *);

/* Enums and arrays */
enum { NO, YES, SEMI };

struct comstruct {
	char *name;
	int (*func)();
	int ptr_param;
	};

#ifdef MAIN_FILE
struct comstruct command[NUM_COMS+1]=
	{
	"NULL",    NULL,      NO,

	"alias",   NULL,      NO,

	"class",   NULL,       NO,
	"endclass",NULL,       NO,
	"public:", NULL,       NO,
	"private:",NULL,       NO,
	"new",     com_new,    NO,
	"dupl",    com_dupl,   NO,
	"delete",  com_delete, NO,
	"promote", com_promote,NO,
	"with",    com_with,   NO,
	"without", com_without,NO,

	"proc",    NULL,      NO,
	"vproc",   NULL,      NO,
	"endproc", NULL,      NO,
	"call",    com_call,  NO,
	"vcall",   com_call,  NO,
	"*",       com_call,  NO,  /* vcall shortcut */
	"return",  com_return,NO,

	"try",           NULL,           NO,
	"tried",         NULL,           NO,
	"catch",         NULL,           NO,
	"catchrun",      NULL,           NO,
	"caught",        NULL,           NO,
	"catchreturn",   com_catchreturn,YES,
	"catchreturnend",com_catchreturn,YES,
	"catchbreak",    com_catchbreak, YES,
	"trap",          com_trap,       NO,
	"throw",         com_throw,      YES,

	"var",     com_var,    NO,
	"set",     com_set,    NO,
	"aset",    com_aset,   NO,
	"afset",   com_aset,   NO,
	"erase",   com_wevar,  NO,
	"wipe",    com_wevar,  NO,
	"arrsize", com_arrsize,NO,

	"if",      com_if,    YES,
	"else",    com_else,  YES,
	"fi",      NULL,      NO,
	"and",     com_bool,  NO,
	"or",      com_bool,  NO,
	"xor",     com_bool,  NO,

	"label",     NULL,          NO,
	"goto",      com_goto,      YES,
	"loop",      com_loopcom,   YES,
	"looping",   com_loopend,   YES,
	"for",       com_for,       YES,
	"next",      com_loopend,   YES,
	"forent",    com_forent,    YES,
	"nextent",   com_loopend,   YES,
	"forelem",   com_forelem,   YES,
	"nextelem",  com_loopend,   YES,
	"while",     com_while,     YES,
	"wend",      com_loopend,   YES,
	"do",        com_do,        NO,
	"whilst",    com_whilst,    YES,
	"switch",    com_switch,    YES,
	"case",      com_case_def,  NO,
	"default",   com_case_def,  NO,
	"switched",  com_loopend,   YES,
	"break",     com_break_cont,YES,
	"continue",  com_break_cont,YES,
	"swcontinue",com_break_cont,YES,

	"select",   com_select,   NO,
	"print",    com_print,    NO,
	"??",       com_print,    NO,  /* print shortcut */
	"printnl",  com_print,    NO, 
	"?",        com_print,    NO,  /* printnl shortcut */
	"stprint",  com_print,    NO,
	"st??",     com_print,    NO,  /* stprint shortcut */
	"stprintnl",com_print,    NO,
	"st?",      com_print,    NO,  /* stprintnl shortcut */
	"input",    com_input,    NO,
	"binput",   com_input,    NO,
	"stinput",  com_input,    NO,
	"stbinput", com_input,    NO,
	"in",       com_in_out,   NO,
	"out",      com_in_out,   NO,
	"open",     com_open,     NO,
	"close",    com_close,    NO,
	"rclose",   com_close,    NO,
	"wclose",   com_close,    NO,
	"wind",     com_wind,     NO,
	"spos",     com_spos,     NO,
	"stat",     com_stat,     NO,
	"lstat",    com_stat,     NO,
	"ioctl",    com_ioctl,    NO,
	"dir",      com_dir,      NO,
	"chdir",    com_chmkdf,   NO,
	"mkdir",    com_chmkdf,   NO,
	"mkfifo",   com_chmkdf,   NO,
	"mkslink",  com_lcm,      NO,
	"mkhlink",  com_lcm,      NO,
	"mv",       com_lcm,      NO,
	"cp",       com_lcm,      NO,
	"rm",       com_rm_attrib,NO,
	"chmod",    com_rm_attrib,NO,
	"chown",    com_rm_attrib,NO,
	"umask",    com_umask,    NO,

	"connect",    com_connect,    NO,
	"listen",     com_listen,     NO,
	"accept",     com_acc_gethost,NO,
	"gethostname",com_acc_gethost,NO,
	"gethostnum", com_acc_gethost,NO,
	"gethostport",com_acc_gethost,NO,
	"gethostinfo",com_acc_gethost,NO,
	"getipnum",   com_getip,      NO,
	"getipname",  com_getip,      NO,

	"addstr",  com_addstr,   NO,
	">>",      com_addstr,   NO,  /* addstr shortcut */
	"strtype", com_strfuncs1,NO,
	"strlen",  com_strfuncs1,NO,
	"uniqstr", com_strfuncs1,NO,
	"nlrm",    com_strfuncs1,NO,
	"chop",    com_strfuncs1,NO,
	"chomp",   com_strfuncs1,NO,
	"revstr",  com_strfuncs1,NO,
	"leftstr", com_strfuncs2,NO,
	"rightstr",com_strfuncs2,NO,
	"matchstr",com_strfuncs2,NO,
	"~",       com_strfuncs2,NO,
	"samestr", com_strfuncs2,NO,
	"diffstr", com_strfuncs2,NO,
	"upperstr",com_ulstr,    NO,
	"lowerstr",com_ulstr,    NO,
	"midstr",  com_midstr,   NO,
	"format",  com_format,   NO,
	"overstr", com_oistr,    NO,
	"inststr", com_oistr,    NO,
	"replstr", com_replstr,  NO,
	"instr",   com_strfuncs3,NO,
	"substr",  com_strfuncs3,NO,
	"<<",      com_strfuncs3,NO,  /* substr shortcut */

	"head",     com_head_tail,NO,
	"tail",     com_head_tail,NO,
	"rhead",    com_head_tail,NO,
	"rtail",    com_head_tail,NO,
	"element" , com_element,  NO,
	"cpelement",com_element,  NO,
	"midlist",  com_midlist,  NO,
	"member",   com_member,   NO,
	"tidy",     com_tidy,     NO,
	"overelem", com_oielem,   NO,
	"instelem", com_oielem,   NO,
	"listsize", com_sizeuniq, NO,
	"uniqlist", com_sizeuniq, NO,
	"samelist", com_sdlist,   NO,
	"difflist", com_sdlist,   NO,
	"sort",     com_sort,     NO,
	"rsort",    com_sort,     NO,
	"nsort",    com_sort,     NO,
	"rnsort",   com_sort,     NO,
	"replace",  com_replace,  NO,
	"join",     com_join_keys,NO,
	"keys",     com_join_keys,NO,
	"split",    com_split,    NO,
	"keysplit", com_split,    NO,

	"eqn",     com_eqn,   NO,
	"not",     com_not,   NO,
	"inc",     com_incdec,NO,
	"+"  ,     com_incdec,NO,  /* Inc shortcut */
	"dec",     com_incdec,NO,
	"-"  ,     com_incdec,NO,  /* Dec shortcut */
	"random",  com_random,NO,
	"min",     com_minmax,NO,
	"max",     com_minmax,NO,
	"sin",     com_trig,  NO,
	"cos",     com_trig,  NO,
	"tan",     com_trig,  NO,
	"asin",    com_trig,  NO,
	"acos",    com_trig,  NO,
	"atan",    com_trig,  NO,
	"exp",     com_maths1,NO,
	"log",     com_maths1,NO,
	"log10",   com_maths1,NO,
	"sqrt",    com_maths1,NO,
	"cbrt",    com_maths1,NO,
	"round",   com_maths1,NO,
	"ceil",    com_maths1,NO,
	"floor",   com_maths1,NO,
	"atan2",   com_maths2,NO,
	"hyp",     com_maths2,NO,
	"frombase",com_maths2,NO,
	"tobase",  com_maths2,NO,

	"sleep", com_sleep,NO,
	"date",  com_date, NO,

	"cls",    com_cls,     NO,
	"locate", com_locate,  NO,
	"oobon",  com_oob,     NO,
	"ooboff", com_oob,     NO,
	"line",   com_line,    NO,
	"circle", com_circle,  NO,
	"ellipse",com_circle,  NO,
	"arc",    com_circle,  NO,
	"getrows",com_termsize,NO,
	"getcols",com_termsize,NO,
	"getsize",com_termsize,NO,

	"signal", com_signal, NO,
	"alarm",  com_alarm,  NO,
	"calarm", com_alarm,  NO,
	"sendsig",com_sendsig,NO,

	"system", com_system,NO,
	"popen",  com_popen, NO,
	"wait",   com_wait,  NO,
	"fork",   com_fork,  NO,
	"exec",   com_exec,  NO,

	"push",     com_push,     NO,
	"pop",      com_pop,      NO,
	"bpush",    com_push,     NO,
	"bpop",     com_pop,      NO,
	"wipestack",com_wipestack,NO,

	"asc",   com_asc,   NO,
	"exit",  com_exit,  NO,
	"stop",  com_exit,  NO,
	"getenv",com_getenv,NO,
	"setenv",com_setenv,NO,
	"crypt", com_crypt, NO,
	};
#else
extern struct comstruct command[];
#endif

enum comenum {
	NOT_COMMAND,

	ALIAS,

	CLASS,
	ENDCLASS,
	PUBLIC,
	PRIVATE,
	NEW,
	DUPL,
	DELETE,
	PROMOTE,
	WITH,
	WITHOUT,

	PROC,
	VPROC_COM, /* VPROC enum element already used in SunOS 5.6 */
	ENDPROC,
	CALL,
	VCALL,
	VCALL_SHORTCUT,
	RETURN,

	TRY,
	TRIED,
	CATCH,
	CATCHRUN,
	CAUGHT,
	CATCHRETURN,    /* DO NOT alter the relative positions of these */
	CATCHRETURNEND, /* two elements */
	CATCHBREAK,
	TRAP,
	THROW,

	VAR,
	SET,
	ASET,
	AFSET,
	ERASE,
	WIPE,
	ARRSIZE,

	IF,
	ELSE,
	FI,
	AND_COM,
	OR_COM,
	XOR_COM,

	LABEL,
	GOTO,
	LOOP,
	LOOPING,
	FOR,
	NEXT,
	FORENT,
	NEXTENT,
	FORELEM,
	NEXTELEM,
	WHILE,
	WEND,
	DO,
	WHILST,
	SWITCH,
	CASE,
	DEFAULT,
	SWITCHED,
	BREAK,
	CONTINUE,
	SWCONTINUE,

	SELECT,
	PRINT,
	PRINT_SHORTCUT,  /* double question mark shortcut */
	PRINTNL,
	PRINTNL_SHORTCUT,  /* question mark shortcut */
	STPRINT,
	STPRINT_SHORTCUT,  /* st?? shortcut */
	STPRINTNL,
	STPRINTNL_SHORTCUT,  /* st? shortcut */
	INPUT,
	BINPUT,
	STINPUT,
	STBINPUT,
	IN,
	OUT,
	OPEN,
	CLOSE,
	RCLOSE,
	WCLOSE,
	WIND,
	SPOS,
	STAT,
	LSTAT,
	IOCTL,
	DIRCOM,
	CHDIR,
	MKDIR,
	MKFIFO,
	MKSLINK,
	MKHLINK,
	MV,
	CP,
	RM,
	CHMOD,
	CHOWN,
	UMASK,

	CONNECT,
	LISTEN,
	ACCEPT,
	GETHOSTNAME,
	GETHOSTNUM,
	GETHOSTPORT,
	GETHOSTINFO,
	GETIPNUM,
	GETIPNAME,

	ADDSTR,
	ADDSTR_SHORTCUT,
	STRTYPE,
	STRLEN,
	UNIQSTR,
	NLRM,
	CHOP,
	CHOMP,
	REVSTR,
	LEFTSTR,
	RIGHTSTR,
	MATCHSTR,
	MATCHSTR_SHORTCUT,
	SAMESTR,
	DIFFSTR,
	UPPERSTR,
	LOWERSTR,
	MIDSTR,
	FORMAT,
	OVERSTR,
	INSTSTR,
	REPLSTR,
	INSTR,
	SUBSTR,
	SUBSTR_SHORTCUT,

	HEAD,
	TAIL,
	RHEAD,
	RTAIL,
	ELEMENT,
	CPELEMENT,
	MIDLIST,
	MEMBER,
	TIDY,
	OVERELEM,
	INSTELEM,
	LISTSIZE,
	UNIQLIST,
	SAMELIST,
	DIFFLIST,
	SORT,
	RSORT,
	NSORT,
	RNSORT,
	REPLACE,
	JOIN,
	KEYS,
	SPLIT,
	KEYSPLIT,

	EQN,
	NOT_COM,
	INC,
	INC_SHORTCUT,
	DEC,
	DEC_SHORTCUT,
	RANDOM,
	MINCOM,
	MAXCOM,
	SIN,
	COS,
	TAN,
	ASIN,
	ACOS,
	ATAN,
	EXP,
	LOG,
	LOG10,
	SQRT,
	CBRT,
	ROUND,
	CEIL,
	FLOOR,
	ATAN2,
	HYP,
	FROMBASE,
	TOBASE,

	SLEEP,
	DATE,

	CLS,
	LOCATE,
	OOBON,
	OOBOFF,
	LINE,
	CIRCLE,
	ELLIPSE,
	ARC,
	GETROWS,
	GETCOLS,
	GETSIZE,

	SIGNAL,
	ALARM,
	CALARM,
	SENDSIG,

	SYSTEM,
	POPEN,
	WAIT,
	FORK,
	EXEC,

	PUSH,
	POP,
	BPUSH,
	BPOP,
	WIPESTACK,

	ASC,
	EXIT,
	STOP,
	GETENV,
	SETENV,
	CRYPT
	};


#ifdef MAIN_FILE
char *error_mesg[NUM_ERRORS]={
	"Ok",
	"INTERNAL ERROR! Please email neil@ogham.demon.co.uk with a bug report",
	"Cannot open file",
	"Cannot stat file",
	"Cannot read link",
	"Cannot open directory",
	"Cannot change directory",
	"Cannot create directory",
	"Cannot create fifo",
	"Cannot create link",
	"Cannot move file",
	"Cannot delete file or directory",
	"Cannot copy a directory",
	"Attempt to copy or move multiple files into one file",
	"Unexpected regular expression",
	"Regular expressions cannot be used in directory names",
	"Cannot change file permissions",
	"Cannot change file ownership",
	"No such user",
	"No such group",
	"Missing bracket(s)",
	"Memory allocation failure",
	"Maximum include file nesting level exceeded",
	"Unterminated string",
	"Syntax error",
	"Invalid argument",
	"Missing ENDPROC",
	"Missing ENDCLASS",
	"Missing FI",
	"Missing CAUGHT",
	"Missing TRIED",
	"Missing WITHOUT",
	"Unexpected command",
	"Command cannot be nested",
	"Duplicate declaration",
	"Invalid variable name",
	"Undefined variable",
	"Invalid subscript",
	"Undefined subscript",
	"Array value not found",
	"Unexpected reference variable",
	"Variable declare-and-set not allowed here",
	"Variable is private",
	"Variable is read only",
	"Cannot negate non numeric variable",
	"Invalid class name",
	"Undefined class",
	"Undefined object",
	"Class cannot inherit itself",
	"Unexpected class specification",
	"The main procedure cannot be a VPROC",
	"A CDD procedure cannot be a VPROC",
	"CDD procedures cannot have arguments",
	"Not in a WITH block",
	"Cannot delete an object currently in context",
	"Object is not derived from the given class",
	"Objects class is not a parent class of class given",
	"Cannot STOP inside a destructor procedure or one within its call stack",
	"Invalid object name",
	"Invalid procedure name",
	"Main procedure missing",
	"Undefined procedure",
	"Procedure is private",
	"Invalid exception",
	"Nested CATCH - CAUGHT blocks",
	"Nested TRY - TRIED blocks",
	"Unterminated loop",
	"Undefined label",
	"Variable is not an array",
	"Array dimensions exceeded",
	"Missing array value in subscript-value pair",
	"Empty expression",
	"Parameter count mismatch",
	"Array dimension mismatch",
	"Reference variables cannot be given an array definition",
	"Cannot reference a non variable",
	"Array must be one dimensional",
	"Cannot wipe variable as an element is being referenced",
	"Value exceeds the current nesting level",
	"Division by zero",
	"Cannot convert to given base",
	"Undefined stream",
	"Stream is for input only",
	"Stream is for output only",
	"Current input stream is closed",
	"Current output stream is closed",
	"Invalid option",
	"Invalid option for stream type",
	"Invalid stream type",
	"Permissions are only allowed with the write or readwrite options",
	"Cannot wind",
	"Invalid format code",
	"Unknown escape code",
	"Missing format argument",
	"Cannot trap the error",
	"Unhandled exception",
	"Cannot CATCHRETURN from this point",
	"Cannot CATCHBREAK at this point",
	"Unknown hostname",
	"Cannot connect to host",
	"Cannot create socket",
	"Cannot bind to port",
	"Cannot listen on port",
	"Cannot accept connection",
	"Unknown signal",
	"Signal handler procedures cannot have arguments",
	"Failed to send signal to process or group",
	"Options are mutually exclusive",
	"Cannot create semaphore",
	"Cannot create pipe",
	"Cannot fork",
	"Execution of process failed",
	"Wait error",
	"Cannot get terminal screen size",
	"Undefined stack",
	"Cannot set enviroment variable",

	"Command, feature or option is unavailable in this build",
	"Command, feature or option is currently unimplemented"
	};
#else
extern char *error_mesg[];
#endif

enum loor_errors {
	OK,
	ERR_INTERNAL,
	ERR_CANT_OPEN_FILE,
	ERR_CANT_STAT_FILE,
	ERR_CANT_READ_LINK,
	ERR_CANT_OPEN_DIR,
	ERR_CANT_CHANGE_DIR,
	ERR_CANT_CREATE_DIR,
	ERR_CANT_CREATE_FIFO,
	ERR_CANT_CREATE_LINK,
	ERR_CANT_MOVE_FILE,
	ERR_CANT_DEL_FILE_OR_DIR,
	ERR_CANT_COPY_DIR,
	ERR_CANT_COPY_MOVE_MULTIPLE_ONTO_ONE,
	ERR_UNEXPECTED_REGEXP,
	ERR_CANT_USE_REGEXP_IN_DIRNAME,
	ERR_CANT_CHANGE_FILE_PERM,
	ERR_CANT_CHANGE_FILE_OWNERSHIP,
	ERR_NO_SUCH_USER,
	ERR_NO_SUCH_GROUP,
	ERR_BRACKETS,
	ERR_MALLOC,
	ERR_MAX_INCLUDE,
	ERR_UNTERMINATED_STRING,
	ERR_SYNTAX,
	ERR_INVALID_ARGUMENT,
	ERR_MISSING_ENDPROC,
	ERR_MISSING_ENDCLASS,
	ERR_MISSING_FI,
	ERR_MISSING_CAUGHT,
	ERR_MISSING_TRIED,
	ERR_MISSING_WITHOUT,
	ERR_UNEXPECTED_COMMAND,
	ERR_COMMAND_CANT_BE_NESTED,
	ERR_DUPLICATE_DECLARATION,
	ERR_INVALID_VARNAME,
	ERR_UNDEFINED_VARIABLE,
	ERR_INVALID_SUBSCRIPT,
	ERR_UNDEFINED_SUBSCRIPT,
	ERR_VALUE_NOT_FOUND,
	ERR_UNEXPECTED_REFVAR,
	ERR_CANT_DECLARE_AND_SET_HERE,
	ERR_PRIVATE_VARIABLE,
	ERR_READONLY_VARIABLE,
	ERR_CANT_NEG_NON_NUM_VAR,
	ERR_INVALID_CLASSNAME,
	ERR_UNDEFINED_CLASS,
	ERR_UNDEFINED_OBJECT,
	ERR_CANT_INHERIT_SELF,
	ERR_UNEXPECTED_CLASS_SPEC,
	ERR_MAIN_IS_VPROC,
	ERR_CDD_IS_VPROC,
	ERR_CDD_ARGS,
	ERR_NOT_IN_WITH_BLOCK,
	ERR_CANT_DEL_OBJ_IN_CONTEXT,
	ERR_OBJ_NOT_DERIVED_FROM_CLASS,
	ERR_OBJ_CLASS_NOT_PARENT,
	ERR_CANT_STOP_IN_DESTRUCTOR,
	ERR_INVALID_OBJNAME,
	ERR_INVALID_PROCNAME,
	ERR_MAIN_PROC_MISSING,
	ERR_UNDEFINED_PROC,
	ERR_PRIVATE_PROC,
	ERR_INVALID_EXCEPTION,
	ERR_NESTED_EXCEPTION_BLOCKS,
	ERR_NESTED_TRY_BLOCKS,
	ERR_UNTERMINATED_LOOP,
	ERR_UNDEFINED_LABEL,
	ERR_VAR_IS_NOT_ARRAY,
	ERR_ARRAY_DIM_EXCEEDED,
	ERR_MISSING_ARRAY_VALUE,
	ERR_EMPTY_EXPRESSION,
	ERR_PARAM_COUNT_MISMATCH,
	ERR_ARRAY_DIM_MISMATCH,
	ERR_REFVAR_CANT_BE_ARRAY,
	ERR_CANT_REF_NON_VAR,
	ERR_ARRAY_NOT_1D,
	ERR_CANT_WIPE_VAR_BEING_REFERENCED,
	ERR_VALUE_EXCEEDS_NESTING,
	ERR_DIVIDE_BY_ZERO,
	ERR_CANT_CONVERT_TO_BASE,
	ERR_UNDEFINED_STREAM,
	ERR_STREAM_INPUT_ONLY,
	ERR_STREAM_OUTPUT_ONLY,
	ERR_INPUT_STREAM_CLOSED,
	ERR_OUTPUT_STREAM_CLOSED,
	ERR_INVALID_OPTION,
	ERR_INVALID_OPTION_FOR_STREAM,
	ERR_INVALID_STREAM_TYPE,
	ERR_PERM_ONLY_WITH_WRITE,
	ERR_CANT_WIND,
	ERR_INVALID_FORMAT_CODE,
	ERR_UNKNOWN_ESC_CODE,
	ERR_MISSING_FORMAT_ARG,
	ERR_CANT_TRAP_ERROR,
	ERR_UNHANDLED_EXCEPTION,
	ERR_CANT_CATCHRETURN,
	ERR_CANT_CATCHBREAK,
	ERR_UNKNOWN_HOST,
	ERR_CONNECT,
	ERR_SOCKET,
	ERR_BIND,
	ERR_LISTEN,
	ERR_ACCEPT,
	ERR_UNKNOWN_SIGNAL,
	ERR_HANDLER_ARGS,
	ERR_FAILED_TO_SEND_SIGNAL,
	ERR_MUTUALLY_EXCLUSIVE,
	ERR_CANT_CREATE_SEMAPHORE,
	ERR_CANT_CREATE_PIPE,
	ERR_CANT_FORK,
	ERR_EXEC_FAILED,
	ERR_WAIT,
	ERR_CANT_GET_TERMSIZE,
	ERR_UNDEFINED_STACK,
	ERR_CANT_SET_ENV_VAR,

	ERR_UNAVAILABLE,
	ERR_UNIMPLEMENTED
	};


#ifdef MAIN_FILE
char *op_string[NUM_OPERATORS+1]={
	"NULL",
	"{",
	"}",
	"(",
	")",
	"<",
	">",
	"!",
	"=",
	"<=",
	">=",
	"!=",
	"%=",
	"*=",
	"=>",
	"+",
	"-",
	"*",
	"/",
	"^",
	"%",
	"&",
	"|",
	"~",
	"<<",
	">>"
	};
#else
extern char *op_string[];
#endif

enum ops {
	/* Single char ops */
	NOT_OPERATOR,
	LEFT_CUR_BRACKET,
	RIGHT_CUR_BRACKET,
	LEFT_RND_BRACKET,
	RIGHT_RND_BRACKET,
	LESS_THAN,
	GREATER_THAN,
	NOT,
	EQUALS,

	/* Multi char ops */
	LESS_THAN_EQUALS,
	GREATER_THAN_EQUALS,
	NOT_EQUAL,
	PCT_EQUALS,
	STAR_EQUALS,
	ASSOCIATIVE_SET,

	/* Maths ops */
	PLUS,
	MINUS,
	MULTIPLY,
	DIVIDE,
	POWER,
	MOD,

	/* Bitwise ops */
	AND,
	OR,
	XOR,
	LEFT_SHIFT,
	RIGHT_SHIFT
	};


enum type_macros {
	NOT_NUMBER,
	INTEGER,
	FLOAT
	};


#define EXCP_TRAP_COUNT 31

/* Errors trapped by exception handlers */
EXTERN
struct {
	char *name;
	int err;
	} err_trapped[EXCP_TRAP_COUNT]
#ifdef MAIN_FILE
	={
	"CANT_OPEN_FILE",       ERR_CANT_OPEN_FILE,
	"CANT_STAT_FILE",       ERR_CANT_STAT_FILE,
	"CANT_READ_LINK",       ERR_CANT_READ_LINK,
	"CANT_OPEN_DIR",        ERR_CANT_OPEN_DIR,
	"CANT_CHANGE_DIR",      ERR_CANT_CHANGE_DIR,
	"CANT_CREATE_DIR",      ERR_CANT_CREATE_DIR,
	"CANT_CREATE_FIFO",     ERR_CANT_CREATE_FIFO,
	"CANT_CREATE_LINK",     ERR_CANT_CREATE_LINK,
	"CANT_MOVE_FILE",       ERR_CANT_MOVE_FILE,
	"CANT_DEL_FILE_OR_DIR", ERR_CANT_DEL_FILE_OR_DIR,
	"CANT_CHANGE_PERM",     ERR_CANT_CHANGE_FILE_PERM,
	"CANT_CHANGE_OWNER",    ERR_CANT_CHANGE_FILE_OWNERSHIP,
	"NO_SUCH_USER",         ERR_NO_SUCH_USER,
	"NO_SUCH_GROUP",        ERR_NO_SUCH_GROUP,
	"OUT_OF_MEMORY",        ERR_MALLOC,
	"DUPLICATE_DECLARATION",ERR_DUPLICATE_DECLARATION,
	"UNDEFINED_SUBSCRIPT",  ERR_UNDEFINED_SUBSCRIPT,
	"VALUE_NOT_FOUND",      ERR_VALUE_NOT_FOUND,
	"DIVIDE_BY_ZERO",       ERR_DIVIDE_BY_ZERO,
	"INPUT_STREAM_CLOSED",  ERR_INPUT_STREAM_CLOSED,
	"OUTPUT_STREAM_CLOSED", ERR_OUTPUT_STREAM_CLOSED,
	"CANT_WIND",            ERR_CANT_WIND,
	"UNKNOWN_HOST",         ERR_UNKNOWN_HOST,
	"CANT_CONNECT",         ERR_CONNECT,
	"CANT_BIND",            ERR_BIND,
	"CANT_LISTEN",          ERR_LISTEN,
	"CANT_ACCEPT",          ERR_ACCEPT,
	"FAILED_TO_SEND_SIGNAL",ERR_FAILED_TO_SEND_SIGNAL,
	"CANT_FORK",            ERR_CANT_FORK,
	"EXEC_FAILED",          ERR_EXEC_FAILED,
	"WAIT_ERROR",           ERR_WAIT
	}
#endif
;

#define NUM_COLS 22
#define COL_RESET "\033[0m"

EXTERN
char *colcode[NUM_COLS]
#ifdef MAIN_FILE
	={
	/* Non-colour codes: reset, bold, underline, blink, reverse, cls */
	"\033[0m", "\033[1m", "\033[4m", "\033[5m", "\033[7m","\033[H\033[J",

	/* Foreground colours: black, red, green, yellow */
	"\033[30m","\033[31m","\033[32m","\033[33m",
	"\033[34m","\033[35m","\033[36m","\033[37m",

	/* Background colours: blue, magenta, turquoise, white */
	"\033[40m","\033[41m","\033[42m","\033[43m",
	"\033[44m","\033[45m","\033[46m","\033[47m"
	}
#endif
;

/* This macro is used to set up the $colcode system variable */
#define COLCOM_LIST \
	"RS OL UL LI RV CL FK FR FG FY FB FM FT FW BK BR BG BY BB BM BT BW"

/* Codes used in a string to produce the colours when prepended with a '~' */
EXTERN
char *colcom[NUM_COLS]
#ifdef MAIN_FILE
	={
	"RS","OL","UL","LI","RV","CL",
	"FK","FR","FG","FY",
	"FB","FM","FT","FW",
	"BK","BR","BG","BY",
	"BB","BM","BT","BW"
	}
#endif
;
