/***************************************************************************
 FILE: objects.c
 LVU : 1.0.0

 DESC:
 This is the low level code for creating classes, and creating , destroying
 and running objects.

 Copyright (C) Neil Robertson 1999
 ***************************************************************************/

#include "globals.h"
		

/*** Parse the class line, create a class and add it to class linked list ***/
int parse_class_line(int pos)
{
CLASS_T c,new;
int len,i,ret;
char *name;

if (!(len=pword[pos].end - pos)) return ERR_SYNTAX;

/* See if we're inheriting anything */
if (len > 1) {
	if (strcasecmp(pword[pos+2].word,"inherits")) return ERR_SYNTAX;
	if (len < 3) return ERR_SYNTAX;
	}
name=pword[pos+1].word;
if (!check_name(name)) return ERR_INVALID_CLASSNAME;

/* See if class already exists */
FOR_ALL_CLASSES(c) 
	if (!strcmp(c->name,name)) return ERR_DUPLICATE_DECLARATION;

/* Create it & add to linked list */
if ((new=create_class(name))==NULL) return ERR_MALLOC;
if (first_class) last_class->next=new;
else first_class=new;
last_class=new;

/* Do inheritance */
if (len > 2) {
	for(i=pos+3;i <= pword[pos].end;++i) {
		if (pword[i].com || pword[i].opcode) return ERR_SYNTAX;
		FOR_ALL_CLASSES(c) {
			if (!strcmp(pword[i].word,c->name)) {
				if (new==c) return ERR_CANT_INHERIT_SELF;
				if ((ret=inherit_class(new,c))!=OK)
					return ret;
				break;
				}
			}
		if (c==NULL) return ERR_UNDEFINED_CLASS;
		}
	}
curr_class=new;
return OK;
}



/*** Create the structure for the class ***/
CLASS_T create_class(char *name)
{
CLASS_T new;

if ((new=(CLASS_T)malloc(sizeof(struct classes)))==NULL) return NULL;
if ((new->name=strdup(name))==NULL) return NULL;

new->inherit_cnt=0;
new->inherit=NULL;
new->var_cnt=0;
new->obj_cnt=0;
new->constructor=NULL;
new->destructor=NULL;
new->duplicator=NULL;
new->var=NULL;
new->next=NULL;

return new;
}



/*** Add a parent class to the childs inherit list ***/
int inherit_class(CLASS_T child, CLASS_T parent)
{
int i;

/* See if we've already inherited class */
for(i=0;i < child->inherit_cnt;++i)
	if (child->inherit[i]==parent) return ERR_DUPLICATE_DECLARATION;

/* Add parent to inherit list */
child->inherit=(CLASS_T *)realloc(child->inherit,sizeof(struct classes *) * child->inherit_cnt+1);
if (child->inherit==NULL) return ERR_MALLOC;
child->inherit[child->inherit_cnt++]=parent;
return OK;
}



/*** Store the name of a variable required to be created when an object is
     created from the class. If cl is set then only parse one variable entry
     since its being called from parse_variable_line() ***/
int parse_class_variable_line(int pos, CLASS_T cl, int single)
{
int p,i,start,end,size;
char *varname,*vn,*cvn;
struct class_vars *new;

if (!single) {
	if ((end=pword[pos].end) - pos < 1) return ERR_SYNTAX;
	start=pos+1;
	}
else { start=pos;  end=pos; }
	
size=sizeof(struct class_vars);

for(p=start;p <= end;++p) {
	/* If variable is a command name or operator then its an error */
	if (pword[p].com || pword[p].opcode) {
		if (pword[p].opcode==AND) return ERR_UNEXPECTED_REFVAR;
		return ERR_INVALID_VARNAME;
		}

	/* Can't declare and set a class variable theres nowhere to store
	   the value (yet, might change this in the future). Have <= end
	   instead of just < end because if its a single then we need to
	   catch follow on opcodes. */
	if (p <= end && pword[p+1].opcode) {
		switch(pword[p+1].opcode) {
			case EQUALS:
			case PCT_EQUALS:
			case STAR_EQUALS:
			return ERR_CANT_DECLARE_AND_SET_HERE;
			}
		}

	/* See if its an array */
	varname=vn=pword[p].word;
	for(varname=vn=pword[p].word;*vn=='@' && *vn;++vn) 
	if (!*vn) return ERR_SYNTAX;

	/* See if its a valid name */
	if (!check_name(vn)) return ERR_INVALID_VARNAME;

	/* See if its already in the class's variable list. Strip off any
	   array '@' beforehand. */
	for(i=0;i < cl->var_cnt;++i) {
		for(cvn=cl->var[i]->name;*cvn=='@';++cvn);
		if (!strcmp(cvn,vn)) 
			return ERR_DUPLICATE_DECLARATION;
		}

	/* Add to list. When I tried to do this only using 1 level of 
	   indirection letting realloc create the structures I got
	   a memory fault with Linux even tho Solaris was happy with it.
	   I still don't understand why. */
	cl->var=(struct class_vars **)realloc(cl->var,sizeof(struct class_vars *) * cl->var_cnt+1);
	if (cl->var==NULL || (new=(struct class_vars *)malloc(size))==NULL)
		return ERR_MALLOC;
	cl->var[i]=new;

	if ((cl->var[cl->var_cnt]->name=strdup(varname))==NULL) 
		return ERR_MALLOC;
	cl->var[cl->var_cnt++]->private=private_member;

	/* Return now if we're only supposed to deal with one variable */
	if (single) return OK;
	}
return OK;
}



/*** Create an object entry ***/
OBJECT_T create_object(CLASS_T class, char *name)
{
OBJECT_T new;
char name2[SCRATCH_SIZE];

if ((new=(OBJECT_T)malloc(sizeof(struct objects)))==NULL) return NULL;

class->obj_cnt++;
if (name) new->name=strdup(name);
else {
	sprintf(name2,"%s:%u",class->name,class->obj_cnt);
	new->name=strdup(name2);
	}
if (new->name==NULL) {
	free(new);  return NULL;
	}

if (first_obj) {
	last_obj->next=new;
	new->prev=last_obj;
	}
else {
	first_obj=new;
	new->prev=NULL;
	}
last_obj=new;

new->next=NULL;
new->class=class;
return new;
}



/* Create the objects variables. This involves walking back up the class
   inheritance list.  */
int create_object_variables(OBJECT_T obj, CLASS_T cl)
{
VAR_T var,v;
int i,ret;
char *vn1,*vn2;

/* Creating the variables for given class */
for(i=0;i < cl->var_cnt;++i) {
	/* See if variable already exists for this object, ie that a class
	   further down the tree has overwritten parent class definition. */
	for(vn1=cl->var[i]->name;*vn1 == '@';++vn1);
	FOR_ALL_VARS(v) {
		for(vn2=v->name;*vn2 == '@';++vn2);
		if (v->owner==obj && !strcmp(vn1,vn2)) break;
		}

	if (v==NULL) {
		if ((var=create_variable(cl->var[i]->name,
		                         obj,
		                         NULL,
		                         cl->var[i]->private,
		                         NO))==NULL)
			return ERR_MALLOC;

		/* Add variable to linked list */
		add_variable(var);
		}
	}

/* Walk back up inheritance list */
for(i=0;i < cl->inherit_cnt; ++i) 
	if ((ret=create_object_variables(obj,cl->inherit[i]))!=OK) return ret;
return OK;
}



/*** Destroy an object and any variables associated with it ***/
int destroy_object(OBJECT_T obj)
{
VAR_T v,next;

/* Delete it from the system array first */
if (delete_array_entry("$objects",obj->name) != OK) return ERR_INTERNAL;

/* Destroy objects variables */
for(v=first_var;v;v=next) {
	next=v->next;
	if (v->owner==obj) {
		remove_variable(v);  destroy_variable(v);
		}
	}

/* Remove object from linked list */
if (obj->next) obj->next->prev=obj->prev;
if (obj->prev) obj->prev->next=obj->next;
if (obj==first_obj) first_obj=obj->next;
if (obj==last_obj) last_obj=obj->prev;

/* Delete it */
free(obj->name);
free(obj);
return OK;
}



/*** Returns YES if child = parent or child is a derived class of parent. ***/
int is_derived_class(CLASS_T child, CLASS_T parent)
{
int i;

if (child == parent) return YES;

for(i=0;i < child->inherit_cnt;++i) 
	if (child->inherit[i] == parent ||
	    is_derived_class(child->inherit[i],parent)) return YES;
return NO;
}



/*** Get the procedure for the class or one of its parents ***/
PROC_T get_class_proc(CLASS_T cl, char *procname)
{
PROC_T proc;
int i;

FOR_ALL_PROCS(proc) {
        if (proc->class == cl && !strcmp(proc->name,procname))
                return proc;
        }
for(i=0;i < cl->inherit_cnt; ++i)
        if (proc=get_class_proc(cl->inherit[i],procname)) return proc;
return NULL;
}



/*** Call all the constructors in a class inheritance tree in reverse order.
     ie start with the parent and work down the tree. If stop_cl is not null
     then the tree walk will stop there for that particular branch. ***/
int call_constructors(CLASS_T cl, CLASS_T stop_cl, OBJECT_T obj)
{
int i,ret;

if (cl==stop_cl) return OK;

for(i=cl->inherit_cnt-1;i >= 0; --i) {
	if ((ret=call_constructors(cl->inherit[i],stop_cl,obj)) != OK)
		return ret;
	}
if (cl->constructor) {
	/* Use '1' as dummy return pc. Could be anything apart from 0 */
	if ((ret=push_pstack(cl->constructor,1,obj)) != OK) return ret;
	if ((ret=mainloop(YES,cl->constructor->start)) != OK) return ret;
	}
return OK;
}



/*** Call all the duplicators in a class inheritance tree in reverse order
     as per constructors ***/
int call_duplicators(CLASS_T cl, OBJECT_T obj)
{
int i,ret;

for(i=cl->inherit_cnt-1;i >= 0; --i) {
	if ((ret=call_duplicators(cl->inherit[i],obj))!=OK)
		return ret;
	}
if (cl->duplicator) {
	if ((ret=push_pstack(cl->duplicator,1,obj)) != OK) return ret;
	if ((ret=mainloop(YES,cl->duplicator->start)) != OK) return ret;
	}
return OK;
}



/*** Call all the destructors. Start with the child and work up the tree. ***/
int call_destructors(CLASS_T cl, OBJECT_T obj)
{
int i,ret;

if (cl->destructor) {
	if ((ret=push_pstack(cl->destructor,1,obj)) != OK) return ret;
	if ((ret=mainloop(YES,cl->destructor->start)) != OK) return ret;
	}
for(i=0;i < cl->inherit_cnt; ++i) {
	if ((ret=call_destructors(cl->inherit[i],obj)) != OK)
		return ret;
	}
return OK;
}
