/***************************************************************************
 FILE: streams.c
 LVU : 1.0.0

 DESC:
 This contains the low level code for creating and destroying streams.
 
 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Open the file and create a stream for it ***/
int create_stream(
	char *filename,  /* For pipes this is the exec string */
	int type,
	int open_mode, 
	mode_t perm,
	int param_fd,
	pid_t pipe_pid,
	STREAM_T *new)
{
#ifdef TIOCGWINSZ
struct winsize ws;
#endif
struct stat fs;
STREAM_T new2;
int fd,flags,ret;

/* Do type action */
switch(type) {
	case ST_CONSOLE: 
	case ST_SOCKET : 
	case ST_LSOCKET: fd=param_fd;  break;

	case ST_FILE:
	/* Convert from Loor stream open_mode to open() flags */
	switch(open_mode) {
		case STM_READ     : flags=O_RDONLY;                      break;
		case STM_WRITE    : flags=O_WRONLY | O_CREAT;            break;
		case STM_APPEND   : flags=O_WRONLY | O_CREAT | O_APPEND; break;
		case STM_READWRITE: flags=O_RDWR | O_CREAT;              break;

		default: return ERR_INTERNAL;
		}

	/* Do an explicit remove() because using the O_TRUNC open() flag
	   truncates the file ok if it exists but any permissions given
	   to open() are ignored and the old file permissions remain. */
	if (open_mode==STM_WRITE) remove(filename);
	if ((fd=open(filename,flags,perm))==-1) return ERR_CANT_OPEN_FILE;
	break;

	case ST_PIPE: break; /* do nothing */

	default: return ERR_INTERNAL;
	}

/* Create stream entry */
if ((new2=(STREAM_T)malloc(sizeof(struct streams)))==NULL) {
	if (param_fd==-1) close(fd);  
	return ERR_MALLOC;
	}
new2->filename=strdup(filename);
new2->fd=fd;
new2->fd2=0;
new2->type=type;
new2->open_mode=open_mode;
new2->block=YES;
new2->stripnl=NO;
new2->read_cnt=0;
new2->write_cnt=0;
new2->telopt_cnt=0;
new2->pid=0;

/* Get terminal size. Do this even on file streams since they could be
   connected to a /dev/tty device for example */
#ifdef TIOCGWINSZ
if ((new2->type != ST_CONSOLE && new2->type != ST_FILE) ||
    ioctl(fd,TIOCGWINSZ,&ws) == -1) {
	new2->rows=0;  new2->cols=0;
	}
else {
	new2->rows=ws.ws_row;  
	new2->cols=ws.ws_col;
	}
#else
new2->rows=0;  
new2->cols=0;
#endif

if (open_mode == STM_APPEND) {
	/* Stat the file to get its size and set pos to file size. */
	if (stat(filename,&fs) == -1) return ERR_CANT_STAT_FILE;
	new2->pos=fs.st_size;
	}
else new2->pos=0;

/* Create a name for the stream using the filename and stream_cnt. 
   If its STDIN, STDOUT or STDERR then leave the name as is. */
switch(type) {
	case ST_CONSOLE:
	new2->name=strdup(filename);  break;

	case ST_PIPE:
	sprintf(text,"PIPE:%u:%u",stream_cnt++,pipe_pid); 
	new2->name=strdup(text);
	break;

	default:
	sprintf(text,"%s:%u",filename,stream_cnt++);
	new2->name=strdup(text); 
	}

/* Add to linked list */
if (first_stream) {
	last_stream->next=new2;
	new2->prev=last_stream;
	}
else {
	first_stream=new2;
	new2->prev=NULL;
	}
new2->next=NULL;
last_stream = *new = new2;

/* Create the array entry */
if ((ret=create_stream_array_entry(new2))!=OK) return ret;

return OK;
}



/*** Remove a stream from the list and destroy it ***/
int destroy_stream(char *name)
{
STREAM_T st;
int ret;

if (isnull(name)) return ERR_UNDEFINED_STREAM;

/* Find stream */
if ((st=get_stream(name)) == NULL)
	return ERR_UNDEFINED_STREAM;

/* Delete stream array entry */
if ((ret=delete_array_entry("$streams",st->name)) != OK) return ret;

/* Delete stream */
if (st==in_stream) {
	in_stream=NULL;
	if (set_var_value("$in_stream",NULL,NULL,YES) != OK) 
		return ERR_INTERNAL;
	}
else 
if (st==out_stream) {
	out_stream=NULL;
	if (set_var_value("$out_stream",NULL,NULL,YES) != OK) 
		return ERR_INTERNAL;
	}

if (st->next) st->next->prev=st->prev;
if (st->prev) st->prev->next=st->next;
if (st==first_stream) first_stream=st->next;
if (st==last_stream) last_stream=st->prev;

/* Don't close actual stdin, out & err */
if (st->type != ST_CONSOLE) {
	if (st->type != ST_PIPE) close(st->fd);
	else {
		/* If its a pipe then the input or output pipes may already 
		   have been closed */
		if (st->open_mode != STM_WRITE) close(st->fd);
		if (st->open_mode != STM_READ) close(st->fd2);

		/* If pid set then we want to reap the child process. Wait
		   until its done. */
		if (st->pid) waitpid(st->pid,NULL,0);
		}
	}

free(st->name);
free(st);

return OK;
}



/*** Add an entry into the stream array ***/
int create_stream_array_entry(STREAM_T st)
{
char *type[]={
	"CONSOLE","FILE","SOCKET","LSOCKET","PIPE"
	};
char *open_modes[]={
	"READ","WRITE","APPEND","READWRITE"
	};
char *blocking[]={
	"NONBLOCK","BLOCK","SEMIBLOCK"
	};
	
char varname[200];


sprintf(text,"%s %s %s %s %u %u %u %s",
              type[st->type],
              open_modes[st->open_mode],
              blocking[st->block],	
              (st->stripnl ? "STRIPNL" : "NOSTRIPNL"),
              st->pos,
              st->read_cnt,
              st->write_cnt,
	      st->filename);

sprintf(varname,"$streams[\"%s\"]",st->name);

if (set_var_value(varname,NULL,text,YES) != OK) return ERR_INTERNAL;
return OK;
}



/*** Get a stream based on its name ***/
STREAM_T get_stream(char *name)
{
STREAM_T st;

if (name==NULL) return NULL;
FOR_ALL_STREAMS(st) 
	if (!strcmp(st->name,name)) return st;
return NULL;
}



/*** Write out to a stream. If its a pipe stream do a wait to reap zombie if
     child process has exited. ***/
void write_stream(STREAM_T st, char *str, int len)
{
if (st->type==ST_PIPE) write(st->fd2,str,len); 
else write(st->fd,str,len); 
}

