/***************************************************************************
 FILE: vars.c
 LVU : 1.0.0

 DESC:
 Low level code for parsing VAR command lines, creating, destroying , getting
 and setting variables.

 Copyright (C) Neil Robertson 1999, 2000
 ***************************************************************************/

#include "globals.h"


/*** Parse the "var a,b ..." line found inside class and procedure
     definitions and also as global definitions ***/
int parse_variable_line(int pos, OBJECT_T owner, PSTACK_T context)
{
CLASS_T cl;
VAR_T new;
int pc,pc2,ret;
char *w,*w2,*endw;

if (!(pword[pos].end - pos)) return ERR_SYNTAX;

for(pc=pos+1;pc <= pword[pos].end;++pc) {
	/* If variable is a command name or operator then its an error */
	if (pword[pc].com || pword[pc].opcode) {
		if (pword[pc].word[0]=='&') return ERR_UNEXPECTED_REFVAR;
		return ERR_INVALID_VARNAME;
		}

	/* See if we're dealing with a class variable defined outside a
	   class-endclass block. eg: var fred::bloggs */
	cl=NULL;
	w=pword[pc].word;
	endw=w+strlen(w);
	for(;w < endw; ++w) {
		if (*w==':') {
			if (*(w+1) != ':') return ERR_INVALID_VARNAME;
			if (context) return ERR_UNEXPECTED_CLASS_SPEC;

			*w='\0';
			FOR_ALL_CLASSES(cl)
				if (!strcmp(cl->name,pword[pc].word)) break;
			*w=':';
			if (cl==NULL) return ERR_UNDEFINED_CLASS;

			/* Don't want next func to see class:: bit at the
			   start of the variable name so frig it */
			w2=pword[pc].word;
			pword[pc].word=w+2;
			ret=parse_class_variable_line(pc,cl,YES);
			pword[pc].word=w2;
			if (ret!=OK) return ret;
			break;
			}
		}
	if (cl) continue;

	/* See if variable already exists or name is reserved */
	if ((ret=check_variable_name(pword[pc].word,owner,context))!=OK)
		return ret;

	/* Create variable and add it to list */
	if ((new=create_variable(pword[pc].word,owner,context,NO,NO))==NULL)
		return ERR_MALLOC;
	add_variable(new);

	/* Check for declare-and-set */
	for(w=pword[pc].word;*w && *w=='@';++w);

	if (pc < pword[pos].end) {
		switch(pword[pc+1].opcode) {
			case EQUALS:
			pc2=pc+2;
			if ((ret=push_rstack_result(&pc2)) != OK) return ret;
			if ((ret=set_var_value(w, context,
			                       rstack_ptr->value,NO)) != OK) 
				return ret;
			pc=pc2-1;
			break;

			/* %= and *= assignments go on to the end of the line 
			   so we won't have any following declarations and we
			   return straight away */
			case PCT_EQUALS: 
			case STAR_EQUALS:
			if (w==pword[pc].word) return ERR_VAR_IS_NOT_ARRAY;

			/* Set temp start of word to miss out '@..' and also
			   set end pos */
			w2=pword[pc].word;
			pword[pc].word=w;
			pword[pc].end=pword[pos].end;
				
			/* Call array set command function */
			ret=com_aset(pword[pc+1].opcode==PCT_EQUALS ? ASET : AFSET,pc);
			pword[pc].word=w2;
			return ret;
			}
		}
	}
return OK;
}



/*** See if variable name has already been used elsewhere ***/
int check_variable_name(
	char *varname,
	OBJECT_T owner,
	PSTACK_T context)
{
VAR_T v;

/* See if its a valid name */
for(;*varname;++varname) if (*varname!='@') break;
if (!*varname) return ERR_SYNTAX;

if (!check_name(varname)) return ERR_INVALID_VARNAME;

/* See if variable already exists */
FOR_ALL_VARS(v) {
	if (v->owner==owner &&
	    !strcmp(varname,v->name) &&
	    (v->context==context || v->context==NULL)) 
		return ERR_DUPLICATE_DECLARATION;
	}
return OK;
}



/*** Create a variable structure. Don't do the linking into the linked list
     here since it may be part of an array ***/
VAR_T create_variable(
	char *varname, 
	OBJECT_T owner,
	PSTACK_T context,
	int private,
	int read_only)
{
VAR_T new;
int dim;

/* See if its an array */
dim=0;
for(;*varname;++varname) 
	if (*varname=='@') ++dim; else break;

/* Alloc mem */
if ((new=(VAR_T)malloc(sizeof(struct vars)))==NULL) return NULL;
if ((new->name=strdup(varname))==NULL) return NULL;
new->dim=dim;
new->value=NULL;
new->refvar=NULL;
new->array=NULL;
new->owner=owner;
new->context=context;
new->private=private;
new->read_only=read_only;
new->prev=NULL;
new->next=NULL;

return new;
}



/*** Add variable to linked list ***/
void add_variable(VAR_T var)
{
if (first_var) {
	last_var->next=var;
	var->prev=last_var;
	}
else {
	first_var=var;
	var->prev=NULL;
	}
var->next=NULL;
last_var=var;
}



/*** Destroy a variable and any array structures associated with it. 
     Reset any forent loop vars. ***/
void destroy_variable(VAR_T var)
{
VAR_T var2,next;
LSTACK_T lst;

FOR_LOOP_STACK(lst) {
	if (lst->forent==var) lst->forent=var->next; 
	if (lst->arr_index==var) lst->arr_index=NULL;
	if (lst->arr_value==var) lst->arr_value=NULL;
	}
free(var->name);
FREE(var->value);

/* Free any array elements */
for(var2=var->array;var2;var2=next) {
	next=var2->next;
	FOR_LOOP_STACK(lst) {
		if (lst->forent==var2) lst->forent=var->next; 
		if (lst->arr_index==var2) lst->arr_index=NULL;
		if (lst->arr_value==var2) lst->arr_value=NULL;
		}
	destroy_variable(var2);
	}
free(var);
}



/*** Remove variable from the main linked list ***/
void remove_variable(VAR_T var)
{
if (var->next) var->next->prev=var->prev;
if (var->prev) var->prev->next=var->next;
if (var==first_var) first_var=var->next;
if (var==last_var) last_var=var->prev;
}



/*** Make variable v_new and exact copy of v_old except for name, owner & 
     context settings. Used to create procedure formal parameters and for
     duplicating objects variables. ***/
int copy_variable(VAR_T v_old, VAR_T v_new, int private)
{
VAR_T v1,v2,new,last;
int ret;

if ((ret=set_string(&v_new->value,v_old->value))!=OK) return ret;
v_new->dim=v_old->dim;
v_new->private=private;
last=v_new;

/* Go through old vars array */
FOR_VAR_ELEMENTS(v_old,v1) {
	if ((new=create_variable(v1->name,NULL,NULL,NO,NO))==NULL)
		return ERR_MALLOC;

	/* Find new spot to plonk array element */
	if (v_new->array==NULL) v_new->array=new;	
	else {
		FOR_VAR_ELEMENTS(v_new,v2)
			if (v2->next==NULL) break;
		v2->next=new;
		}
	new->value=strdup(v1->value);
	new->next=NULL;
	new->prev=last;
	last=new;
	if (v1->array) {
		if ((ret=copy_variable(v1,new,private))!=OK) return ret;
		}
	}	
return OK;
}



/*** Get the value of a variable. If we did a reverse lookup and the value
     is not found return an empty string rather than erroring. ***/
int get_var_value(char *name, PSTACK_T context)
{
#ifndef PATH_MAX
#define PATH_MAX 256
#endif
VAR_T var;
PSTACK_T context2;
OBJECT_T obj;
struct timeval tv;
char *substart,*name1,timestr[10],dir[PATH_MAX+1];
int ret,rev_lookup;

if (name==NULL) return ERR_INVALID_ARGUMENT;

/* Check for the time variables who's value is not stored in their value
   field. These are guaranteed to exist so we don't need to do all the 
   getset_var calls first. This speeds up the return of their values but
   slows it down for other system variables. Oh well. */
if (name[0]=='$') {
	name1=name+1;
	gettimeofday(&tv,NULL);
	if (!strcmp(name1,"time")) {
		sprintf(timestr,"%u",tv.tv_sec);
		return push_rstack(timestr);
		}
	if (!strcmp(name1,"utime")) {
		sprintf(timestr,"%u",tv.tv_usec);
		return push_rstack(timestr);
		}
	if (!strcmp(name1,"dir")) {
		getwd(dir);
		return push_rstack(dir);
		}
	if (!strcmp(name1,"this")) {
		/* Go back up proc call stack to find the current object 
		   context we're in */
		for(context2=context;context2;context2=context2->prev)	
			if (context2->obj) 
				return push_rstack(context2->obj->name);	
		return push_rstack("");
		}
	if (!strcmp(name1,"self")) {
		/* Go back up proc call stack to find the initial object 
		   context we're in */
		for(context2=context,obj=NULL;context2;context2=context2->prev)	
			if (context2->obj) obj=context2->obj;
		if (obj) return push_rstack(obj->name);
		return push_rstack("");
		}
	}

if ((ret=get_var_main(name,context,&var,&substart))!=OK) return ret;

rev_lookup=0;
if (substart) 
	ret=getset_var_element(var,context,&var,NULL,substart,NO,&rev_lookup);

if (rev_lookup) {
	if (ret!=OK) {
		if (ret==ERR_VALUE_NOT_FOUND) return push_rstack(NULL);
		return ret;
		}
	return push_rstack(var->name);
	}
else if (ret!=OK) return ret;

return push_rstack(var->value);
}



/*** Find the main variable given by the name. If the variable is an
     array element get_var_element must be called after to get the pointer
     to the actual element itself. ***/
int get_var_main(
	char *name,
	PSTACK_T context,
	VAR_T *var,
	char **substart)
{
VAR_T v;
OBJECT_T obj;
char name2[SCRATCH_SIZE];
char *s,*name3,c;
int ret,brackets;

/* This is required in case the variable name is hard coded into the 
   source code for this program as trying to change any of the actual
   name would cause a seggy */
strcpy(name2,name);

*var=NULL;
*substart=NULL;

/* First see if its internal to an object. We go backwards here so that
   mutiple references can be put together. Eg: abc.def->ghi */
obj=NULL;
brackets=0;
s=name2+strlen(name2)-1;;

while(s >= name2) {
	if (*s==']') brackets++;
	else
	if (*s=='[') brackets--;

	if (!brackets && (*s=='.' || *s=='>')) {
		if (*s=='>') {
			--s;
			if (*s!='-') return ERR_INVALID_VARNAME;
			}
		if (s <= name2 && *s!='.') return ERR_INVALID_VARNAME;

		/* Find the object */
		c=*s;  
		*s='\0';
		if (c=='-') {
			/* A variable contains the object name. eg: a->b */
			if ((ret=get_var_value(name2,context))!=OK) {
				*s=c;  return ret;
				}
			if (rstack_ptr->value==NULL) 
				return ERR_INVALID_ARGUMENT;

			FOR_ALL_OBJECTS(obj) 
				if (!strcmp(obj->name,rstack_ptr->value)) break;
			}
		else {
			/* See if its ".var" in which case we should be in
			   a WITH block */
			if (!name2[0]) {
				if (wstack_ptr==NULL)
					return ERR_NOT_IN_WITH_BLOCK;
				obj=wstack_ptr->obj;
				}
			else {
				/* Object name hard coded into code. */
				FOR_ALL_OBJECTS(obj) 
					if (!strcmp(obj->name,name2)) break;
				}
			}
		*s=c;
		if (obj==NULL) return ERR_UNDEFINED_OBJECT;
		break;
		}
	--s;
	}

/* Put s back to the first character */
if (s < name2) s=name2;

/* Now find the variable itself. If we're refering to a specific object
   variable we don't care about the stack context so set it to NULL */
if (obj) context=NULL;

/* Get start of var name */
if (*s=='.') name3=s+1;
else if (*s=='-') name3=s+2;
else name3=name2;

/* Terminate at first array bracket */
while(*s && *s!='[') ++s;
c='\0';
if (*s=='[') {  c=*s;  *s='\0';  }

FOR_ALL_VARS(v) {
	if (v->owner==obj && 
	    (v->context==context || v->context==NULL) &&
	    !strcmp(v->name,name3)) break;
	}
*s=c;
if (v==NULL) return ERR_UNDEFINED_VARIABLE;

/* If we're a reference variable find variable we refer to. This may be
   several levels back if reference variable in turn is referencing another
   reference variable */
while(v->refvar) v=v->refvar;

/* See if variable is private. If it is and we're not in a procedure of
   the class or a child class to which the variable belongs to then we
   can't access it. */
if (obj && 
    v->private && 
    !is_derived_class(obj->class,pstack_ptr->proc->class))
		return ERR_PRIVATE_VARIABLE;

if (!v->dim && c=='[') return ERR_VAR_IS_NOT_ARRAY;

*var=v;  
if (*s) *substart=s;
return OK;
}



/*** Find the sub variable (array element) that matches the array passed.
     This function DOES NOT check to see if a variable is a reference
     variable. It is assumed get_var_main() has done this already and
     the var pointer passed here is to the original referenced variable. ***/
int getset_var_element(
	VAR_T var, 
	PSTACK_T context,
	VAR_T *retvar, 
	char *set_value,
	char *subscripts,
	int set,
	int *rev_lookup)
{
VAR_T ovar,var2,new;
char *s,*s2,*elem_value;
char element[SCRATCH_SIZE];
int brackets,index,ret,icnt;
int treat_as_index;
int dim;

dim=0;
ovar=var;
s=subscripts;
while(*s) {
	if (++dim > ovar->dim) return ERR_ARRAY_DIM_EXCEEDED;
	s2=s;

	/* Get the individual element name */
	if (*s!='[') return ERR_INVALID_ARGUMENT;
	++s;
	brackets=1;
	while(*s) {
		if (*s=='[') ++brackets;
		if (*s==']') {
			--brackets;
			if (!brackets) break;
			}
		++s;
		}
	if (!*s) return ERR_SYNTAX;

	strncpy(element,s2+1,(s-s2-1));
	element[s-s2-1]='\0';
	++s;

	/* Get element type, ie: number, string or variable name and get
	   value */
	index=0;
	elem_value=NULL;
	*rev_lookup=0;
	treat_as_index=0;

	if (element[0]=='?') {
		/* Do a reverse lookup , ie search on subscript value
		   rather than name. */
		*rev_lookup=1;  strcpy(element,element+1);
		}
	if (element[0]=='"') {
		elem_value=element+1;  
		element[s-s2-3-*rev_lookup]='\0'; 
		if (!*elem_value) return ERR_INVALID_SUBSCRIPT;
		}
	else {
		if (element[0]=='#') {
			/* Force the value in a variable to be treated as
			   an index number rather than the subscript value */ 
			treat_as_index=1;  strcpy(element,element+1);
			}

		switch(number_type(element)) {
			case INTEGER:
			if (treat_as_index) {
				if ((index=atoi(element)) < 1) 
					return ERR_INVALID_SUBSCRIPT;
				}
			else elem_value=element;
			break;

			case FLOAT: 
			if (treat_as_index) return ERR_INVALID_SUBSCRIPT;
			elem_value=element;
			break;

			case NOT_NUMBER:
			if ((ret=get_var_value(element,context))!=OK) 
				return ret;
			if (isnull(rstack_ptr->value)) 
				return ERR_INVALID_SUBSCRIPT;
			elem_value=rstack_ptr->value;

			if (treat_as_index) {
				if (number_type(elem_value) != INTEGER ||
				    (index=atoi(elem_value)) < 1) 
					return ERR_INVALID_SUBSCRIPT;
				}
			break;

			default:
			return ERR_INTERNAL;
			}
		}

	/* Find the node for the given level */
	icnt=1;
	FOR_VAR_ELEMENTS(var,var2) {
		if (icnt++==index) break;
		if (*rev_lookup) {
			if ((var2->value==NULL && !element[0]) ||
			    (var2->value && !strcmp(var2->value,elem_value))) 
				break;
			}
		else
		if (!index && 
		    elem_value && 
		    !strcmp(var2->name,elem_value)) break;
		}

	/* If GETing the element and it doesn't exist return error */
	if (!set) {
		if ((*retvar=var2)==NULL) {
			if (*rev_lookup) return ERR_VALUE_NOT_FOUND;
			return ERR_UNDEFINED_SUBSCRIPT;
			}
		var=var2;
		continue;
		}

	/* SET requires us to create the variable element. Cannot create one
	   based on an index or a reverse lookup. */
	if (var2==NULL) {
		if (index) return ERR_UNDEFINED_SUBSCRIPT;
		if (*rev_lookup) return ERR_VALUE_NOT_FOUND;
		if ((new=create_variable(elem_value,NULL,NULL,NO,NO))==NULL)
			return ERR_MALLOC;

		new->dim=ovar->dim - dim;

		if (var->array==NULL) var->array=new;
		else {
			FOR_VAR_ELEMENTS(var,var2)
				if (var2->next==NULL) break;
			var2->next=new;
			new->prev=var2; 
			}
		var2=new;
		}
	var=var2;
	}
if (set) {
	if (retvar) *retvar=var;
	return set_string(&var->value,set_value);
	}
if ((*retvar=var)==NULL) return ERR_UNDEFINED_SUBSCRIPT;
return OK;
}



/*** Set the value of a variable ***/
int set_var_value(
	char *name, 
	PSTACK_T context, 
	char *value, 
	int force)
{
VAR_T var;
char *substart;
int ret,dummy;

if (name==NULL) return ERR_INVALID_ARGUMENT;

if ((ret=get_var_main(name,context,&var,&substart))!=OK) return ret;
if (!force && var->read_only) return ERR_READONLY_VARIABLE;

/* If no subscript specified set the main variable */
if (substart==NULL) {
	/* Check for internal variables */
	if (!strcmp(var->name,"$ES")) set_string(&element_seps,value);	
	if (!strcmp(var->name,"$empty")) set_string(&empty_stack_marker,value);

	return set_string(&var->value,value);
	}

/* Attempt to set or create subscript element */
return getset_var_element(var,context,NULL,value,substart,YES,&dummy);
}



/*** Remove an entry from the given array ***/
int delete_array_entry(char *array_name, char *element_name)
{
VAR_T v1,v2,v3;
LSTACK_T lst;

/* Get main array variable */
FOR_ALL_VARS(v1)
        if (v1->context==NULL && !strcmp(v1->name,array_name)) break;
if (v1==NULL) return ERR_INTERNAL;

/* Get array entry */
FOR_VAR_ELEMENTS(v1,v2)
        if (!strcmp(v2->name,element_name)) break;
if (v2==NULL) return ERR_INTERNAL;

/* Check var isn't being referenced */
FOR_ALL_VARS(v3)
        if (v3->refvar==v2) return ERR_CANT_WIPE_VAR_BEING_REFERENCED;

/* If being referenced in forent loop then alter loop var */
FOR_LOOP_STACK(lst) {
        if (lst->forent==v2) lst->forent=v2->next;
        if (lst->arr_index==v2) lst->arr_index=NULL;
        if (lst->arr_value==v2) lst->arr_index=NULL;
        }

/* Reset initial array pointer if appropriate */
if (v2 == v1->array) v1->array=v2->next;

/* Delete array entry var */
remove_variable(v2);
destroy_variable(v2);
return OK;
}

